#' @name clustering
#' @rdname clustering
#' @aliases clusteringTask1 computeClusters1 computeClusters2
#'
#' @title Two-stage clustering, withing one task (see \code{claws()})
#'
#' @description \code{clusteringTask1()} runs one full stage-1 task, which consists in
#'   iterated stage 1 clustering (on nb_curves / ntasks energy contributions, computed
#'   through discrete wavelets coefficients).
#'   \code{clusteringTask2()} runs a full stage-2 task, which consists in synchrones
#'   and then WER distances computations, before applying the clustering algorithm.
#'   \code{computeClusters1()} and \code{computeClusters2()} correspond to the atomic
#'   clustering procedures respectively for stage 1 and 2. The former applies the
#'   clustering algorithm (PAM) on a contributions matrix, while the latter clusters
#'   a chunk of series inside one task (~max nb_series_per_chunk)
#'
#' @param indices Range of series indices to cluster in parallel (initial data)
#' @param getContribs Function to retrieve contributions from initial series indices:
#'   \code{getContribs(indices)} outpus a contributions matrix
#' @param contribs matrix of contributions (e.g. output of \code{curvesToContribs()})
#' @inheritParams computeSynchrones
#' @inheritParams claws
#'
#' @return For \code{clusteringTask1()} and \code{computeClusters1()}, the indices of the
#'   computed (K1) medoids. Indices are irrelevant for stage 2 clustering, thus
#'   \code{computeClusters2()} outputs a big.matrix of medoids
#'   (of size limited by nb_series_per_chunk)
NULL

#' @rdname clustering
#' @export
clusteringTask1 = function(
	indices, getContribs, K1, nb_series_per_chunk, ncores_clust=1, verbose=FALSE, parll=TRUE)
{
	if (verbose)
		cat(paste("*** Clustering task on ",length(indices)," lines\n", sep=""))

	wrapComputeClusters1 = function(inds) {
		if (parll)
			require("epclust", quietly=TRUE)
		if (verbose)
			cat(paste("   computeClusters1() on ",length(inds)," lines\n", sep=""))
		inds[ computeClusters1(getContribs(inds), K1) ]
	}

	if (parll)
	{
		cl = parallel::makeCluster(ncores_clust)
		parallel::clusterExport(cl, varlist=c("getContribs","K1","verbose"), envir=environment())
	}
	while (length(indices) > K1)
	{
		indices_workers = .spreadIndices(indices, nb_series_per_chunk)
		if (parll)
			indices = unlist( parallel::parLapply(cl, indices_workers, wrapComputeClusters1) )
		else
			indices = unlist( lapply(indices_workers, wrapComputeClusters1) )
	}
	if (parll)
		parallel::stopCluster(cl)

	indices #medoids
}

#' @rdname clustering
#' @export
clusteringTask2 = function(medoids, K2,
	getRefSeries, nb_ref_curves, nb_series_per_chunk, ncores_clust=1,verbose=FALSE,parll=TRUE)
{
	if (nrow(medoids) <= K2)
		return (medoids)
	synchrones = computeSynchrones(medoids,
		getRefSeries, nb_ref_curves, nb_series_per_chunk, ncores_clust, verbose, parll)
	distances = computeWerDists(synchrones, ncores_clust, verbose, parll)
	# PAM in package 'cluster' cannot take big.matrix in input: need to cast it
	mat_dists = matrix(nrow=K1, ncol=K1)
	for (i in seq_len(K1))
		mat_dists[i,] = distances[i,]
	medoids[ computeClusters2(mat_dists,K2), ]
}

#' @rdname clustering
#' @export
computeClusters1 = function(contribs, K1)
	cluster::pam(contribs, K1, diss=FALSE)$id.med

#' @rdname clustering
#' @export
computeClusters2 = function(distances, K2)
	cluster::pam(distances, K2, diss=TRUE)$id.med

#' computeSynchrones
#'
#' Compute the synchrones curves (sum of clusters elements) from a matrix of medoids,
#' using L2 distances.
#'
#' @param medoids big.matrix of medoids (curves of same length as initial series)
#' @param getRefSeries Function to retrieve initial series (e.g. in stage 2 after series
#'   have been replaced by stage-1 medoids)
#' @param nb_ref_curves How many reference series? (This number is known at this stage)
#' @inheritParams claws
#'
#' @return A big.matrix of size K1 x L where L = data_length
#'
#' @export
computeSynchrones = function(medoids, getRefSeries,
	nb_ref_curves, nb_series_per_chunk, ncores_clust=1,verbose=FALSE,parll=TRUE)
{
	computeSynchronesChunk = function(indices)
	{
		if (verbose)
			cat(paste("--- Compute synchrones for ",length(indices)," lines\n", sep=""))
		ref_series = getRefSeries(indices)
		#get medoids indices for this chunk of series
		for (i in seq_len(nrow(ref_series)))
		{
			j = which.min( rowSums( sweep(medoids, 2, ref_series[i,], '-')^2 ) )
			if (parll)
				synchronicity::lock(m)
			synchrones[j,] = synchrones[j,] + ref_series[i,]
			counts[j,1] = counts[j,1] + 1
			if (parll)
				synchronicity::unlock(m)
		}
	}

	K = nrow(medoids)
	# Use bigmemory (shared==TRUE by default) + synchronicity to fill synchrones in //
	# TODO: if size > RAM (not our case), use file-backed big.matrix
	synchrones = bigmemory::big.matrix(nrow=K,ncol=ncol(medoids),type="double",init=0.)
	counts = bigmemory::big.matrix(nrow=K,ncol=1,type="double",init=0)
	# synchronicity is only for Linux & MacOS; on Windows: run sequentially
	parll = (requireNamespace("synchronicity",quietly=TRUE)
		&& parll && Sys.info()['sysname'] != "Windows")
	if (parll)
		m <- synchronicity::boost.mutex()

	if (parll)
	{
		cl = parallel::makeCluster(ncores_clust)
		parallel::clusterExport(cl,
			varlist=c("synchrones","counts","verbose","medoids","getRefSeries"),
			envir=environment())
	}

	indices_workers = .spreadIndices(seq_len(nb_ref_curves), nb_series_per_chunk)
	ignored <-
		if (parll)
			parallel::parLapply(indices_workers, computeSynchronesChunk)
		else
			lapply(indices_workers, computeSynchronesChunk)

	if (parll)
		parallel::stopCluster(cl)

	#TODO: can we avoid this loop? ( synchrones = sweep(synchrones, 1, counts, '/') )
	for (i in seq_len(K))
		synchrones[i,] = synchrones[i,] / counts[i,1]
	#NOTE: odds for some clusters to be empty? (when series already come from stage 2)
	#      ...maybe; but let's hope resulting K1' be still quite bigger than K2
	noNA_rows = sapply(seq_len(K), function(i) all(!is.nan(synchrones[i,])))
	if (all(noNA_rows))
		return (synchrones)
	# Else: some clusters are empty, need to slice synchrones
	synchrones[noNA_rows,]
}

#' computeWerDists
#'
#' Compute the WER distances between the synchrones curves (in rows), which are
#' returned (e.g.) by \code{computeSynchrones()}
#'
#' @param synchrones A big.matrix of synchrones, in rows. The series have same length
#'   as the series in the initial dataset
#' @inheritParams claws
#'
#' @return A big.matrix of size K1 x K1
#'
#' @export
computeWerDists = function(synchrones, ncores_clust=1,verbose=FALSE,parll=TRUE)
{



#TODO: re-organize to call computeWerDist(x,y) [C] (in //?) from two indices + big.matrix


	n <- nrow(synchrones)
	delta <- ncol(synchrones)
	#TODO: automatic tune of all these parameters ? (for other users)
	nvoice   <- 4
	# noctave = 2^13 = 8192 half hours ~ 180 days ; ~log2(ncol(synchrones))
	noctave = 13
	# 4 here represent 2^5 = 32 half-hours ~ 1 day
	#NOTE: default scalevector == 2^(0:(noctave * nvoice) / nvoice) * s0 (?)
	scalevector  <- 2^(4:(noctave * nvoice) / nvoice + 1)
	#condition: ( log2(s0*w0/(2*pi)) - 1 ) * nvoice + 1.5 >= 1
	s0=2
	w0=2*pi
	scaled=FALSE
	s0log = as.integer( (log2( s0*w0/(2*pi) ) - 1) * nvoice + 1.5 )
	totnoct = noctave + as.integer(s0log/nvoice) + 1

	computeCWT = function(i)
	{
		if (verbose)
			cat(paste("+++ Compute Rwave::cwt() on serie ",i,"\n", sep=""))
		ts <- scale(ts(synchrones[i,]), center=TRUE, scale=scaled)
		totts.cwt = Rwave::cwt(ts, totnoct, nvoice, w0, plot=FALSE)
		ts.cwt = totts.cwt[,s0log:(s0log+noctave*nvoice)]
		#Normalization
		sqs <- sqrt(2^(0:(noctave*nvoice)/nvoice)*s0)
		sqres <- sweep(ts.cwt,2,sqs,'*')
		sqres / max(Mod(sqres))
	}

	if (parll)
	{
		cl = parallel::makeCluster(ncores_clust)
		parallel::clusterExport(cl,
			varlist=c("synchrones","totnoct","nvoice","w0","s0log","noctave","s0","verbose"),
			envir=environment())
	}

	# list of CWT from synchrones
	# TODO: fit in RAM, OK? If not, 2 options: serialize, compute individual distances
	Xcwt4 <-
		if (parll)
			parallel::parLapply(cl, seq_len(n), computeCWT)
		else
			lapply(seq_len(n), computeCWT)

	if (parll)
		parallel::stopCluster(cl)

	Xwer_dist <- bigmemory::big.matrix(nrow=n, ncol=n, type="double")
	fcoefs = rep(1/3, 3) #moving average on 3 values (TODO: very slow! correct?!)
	if (verbose)
		cat("*** Compute WER distances from CWT\n")

	#TODO: computeDistances(i,j), et répartir les n(n-1)/2 couples d'indices
	#là c'est trop déséquilibré

	computeDistancesLineI = function(i)
	{
		if (verbose)
			cat(paste("   Line ",i,"\n", sep=""))
		for (j in (i+1):n)
		{
			#TODO: 'circular=TRUE' is wrong, should just take values on the sides; to rewrite in C
			num <- filter(Mod(Xcwt4[[i]] * Conj(Xcwt4[[j]])), fcoefs, circular=TRUE)
			WX <- filter(Mod(Xcwt4[[i]] * Conj(Xcwt4[[i]])), fcoefs, circular=TRUE)
			WY <- filter(Mod(Xcwt4[[j]] * Conj(Xcwt4[[j]])), fcoefs, circular=TRUE)
			wer2 <- sum(colSums(num)^2) / sum( sum(colSums(WX) * colSums(WY)) )
			if (parll)
				synchronicity::lock(m)
			Xwer_dist[i,j] <- sqrt(delta * ncol(Xcwt4[[1]]) * (1 - wer2))
			Xwer_dist[j,i] <- Xwer_dist[i,j]
			if (parll)
				synchronicity::unlock(m)
		}
		Xwer_dist[i,i] = 0.
	}

	parll = (requireNamespace("synchronicity",quietly=TRUE)
		&& parll && Sys.info()['sysname'] != "Windows")
	if (parll)
		m <- synchronicity::boost.mutex()

	ignored <-
		if (parll)
		{
			parallel::mclapply(seq_len(n-1), computeDistancesLineI,
				mc.cores=ncores_clust, mc.allow.recursive=FALSE)
		}
		else
			lapply(seq_len(n-1), computeDistancesLineI)
	Xwer_dist[n,n] = 0.
	Xwer_dist
}

# Helper function to divide indices into balanced sets
.spreadIndices = function(indices, nb_per_chunk)
{
	L = length(indices)
	nb_workers = floor( L / nb_per_chunk )
	if (nb_workers == 0)
	{
		# L < nb_series_per_chunk, simple case
		indices_workers = list(indices)
	}
	else
	{
		indices_workers = lapply( seq_len(nb_workers), function(i)
			indices[(nb_per_chunk*(i-1)+1):(nb_per_chunk*i)] )
		# Spread the remaining load among the workers
		rem = L %% nb_per_chunk
		while (rem > 0)
		{
			index = rem%%nb_workers + 1
			indices_workers[[index]] = c(indices_workers[[index]], indices[L-rem+1])
			rem = rem - 1
		}
	}
	indices_workers
}
